METHOD_CONNECTION_START = 10
METHOD_CONNECTION_START_OK = 11
METHOD_CONNECTION_SECURE = 20
METHOD_CONNECTION_SECURE_OK = 21
METHOD_CONNECTION_TUNE = 30
METHOD_CONNECTION_TUNE_OK = 31
METHOD_CONNECTION_OPEN = 40
METHOD_CONNECTION_OPEN_OK = 41
METHOD_CONNECTION_CLOSE = 50
METHOD_CONNECTION_CLOSE_OK = 51

METHOD_CHANNEL_OPEN = 10
METHOD_CHANNEL_OPEN_OK = 11
METHOD_CHANNEL_FLOW = 20
METHOD_CHANNEL_FLOW_OK = 21
METHOD_CHANNEL_CLOSE = 40
METHOD_CHANNEL_CLOSE_OK = 41

METHOD_EXCHANGE_DECLARE = 10
METHOD_EXCHANGE_DECLARE_OK = 11
METHOD_EXCHANGE_DELETE = 20
METHOD_EXCHANGE_DELETE_OK = 21
METHOD_EXCHANGE_BIND = 30  # rabbitmq extension
METHOD_EXCHANGE_BIND_OK = 31  # rabbitmq extension

METHOD_QUEUE_DECLARE = 10
METHOD_QUEUE_DECLARE_OK = 11
METHOD_QUEUE_BIND = 20
METHOD_QUEUE_BIND_OK = 21
METHOD_QUEUE_PURGE = 30
METHOD_QUEUE_PURGE_OK = 31
METHOD_QUEUE_DELETE = 40
METHOD_QUEUE_DELETE_OK = 41
METHOD_QUEUE_UNBIND = 50
METHOD_QUEUE_UNBIND_OK = 51

METHOD_BASIC_QOS = 10
METHOD_BASIC_QOS_OK = 11
METHOD_BASIC_CONSUME = 20
METHOD_BASIC_CONSUME_OK = 21
METHOD_BASIC_CANCEL = 30
METHOD_BASIC_CANCEL_OK = 31
METHOD_BASIC_PUBLISH = 40
METHOD_BASIC_RETURN_ = 50
METHOD_BASIC_DELIVER = 60
METHOD_BASIC_GET = 70
METHOD_BASIC_GET_OK = 71
METHOD_BASIC_GET_EMPTY = 72
METHOD_BASIC_ACK = 80
METHOD_BASIC_REJECT = 90
METHOD_BASIC_RECOVER_ASYNC = 100
METHOD_BASIC_RECOVER = 110
METHOD_BASIC_RECOVER_OK = 111
METHOD_BASIC_NACK = 120  # RabbitMQ extension

# whole class is a RabbitMQ extension
METHOD_CONFIRM_SELECT = 10
METHOD_CONFIRM_SELECT_OK = 11



FRAME_TYPE_BODY = 3


class BodyFrame:
    frame_type = FRAME_TYPE_BODY
    __slots__ = 'channel_id', 'payload_size', 'payload'

    def __init__(self, channel_id: int, payload_size: int, payload: bytes):
        self.channel_id = channel_id
        self.payload_size = payload_size
        self.payload = payload

    def __str__(self):
        return 'Body frame on channel {} with {} bytes of data'.format(self.channel_id, self.payload_size)



from typing import Union, Set

from ammoo.wire.frames import Frame
from ammoo.wire.low.exceptions import InvalidParse
from ammoo.wire.class_and_method import pretty_format_cam, cam_long_uint
from ammoo.wire.typing import ReplyCode
from ammoo.wire.frames.method.connection import ConnectionCloseParameters


class ConnectionClosed(Exception):
    """Superclass for connection lost + server and client closed exceptions. Not instantiated directly."""

    def __init__(self):
        raise Exception('Do not call me')


class ConnectionLost(ConnectionClosed):
    """This means the broker closed the connection without a connection.close method frame, or we just lost network
    connectivity."""

    def __init__(self, exc: Exception):
        self.exc = exc  # type: Exception

    def __str__(self):
        return 'Connection to server was lost: {}'.format(self.exc)


class ServerClosedConnection(ConnectionClosed):
    """Server sent a connection.close method frame"""

    def __init__(self, parameters: ConnectionCloseParameters):
        self.reply_code = parameters.reply_code  # type: ReplyCode
        self.reply_text = parameters.reply_text  # type: str
        self.class_id = parameters.class_id  # type: int
        self.method_id = parameters.method_id  # type: int

    def __str__(self):
        if self.class_id != 0 and self.method_id != 0:
            cam_text = ' because of method {}'.format(pretty_format_cam(cam_long_uint(self.class_id, self.method_id)))
        else:
            cam_text = ''
        return 'Server closed connection with code {}{}: {}'.format(self.reply_code, cam_text, self.reply_text)


class ClientClosedConnection(ConnectionClosed):
    pass


class ClientClosedConnectionOK(ClientClosedConnection):
    def __init__(self):
        pass

    def __str__(self):
        return 'Client closed connection without errors'


class ClientClosedConnectionError(ClientClosedConnection):
    """Use this class only to wrap exceptions we did not expect to happen. Ie. we do not know why self.exc was raised"""

    def __init__(self, exc: Exception):
        self.exc = exc  # type: Exception

    def __str__(self):
        return 'Client closed connection because of an unhandled exception: {!r}'.format(self.exc)


class ServerAuthMechanismsNotSupportedError(ClientClosedConnection):
    def __init__(self):
        pass

    def __str__(self):
        return 'Client closed connection because none of the server\'s auth mechanisms are supported'


class AuthenticationError(ClientClosedConnection):
    """Raised by authentication mechanism classes in build_start_ok_response() or challenge() when they're not able
    to authenticate with the server. May be subclassed to provide a more specific exception"""

    def __init__(self, mechanism: str):
        self.mechanism = mechanism

    def __str__(self):
        return 'Client closed connection because it could not authenticate with the server using the {} mechanism'.format(
            self.mechanism
        )


class NoChallengesAuthenticationError(AuthenticationError):
    def __str__(self):
        return (
            'Client closed connection because the server sent a connection.secure method with a challenge, which '
            'authentication mechanism {} does not support'
        ).format(self.mechanism)


class ServerLocalesNotSupportedError(ClientClosedConnection):
    def __init__(self, server_locales: Set[str]):
        self.server_locales = server_locales

    def __str__(self) -> str:
        return 'Client closed connection because none of the server\'s locales are supported: {}'.format(
            ', '.join(self.server_locales)
        )


class HeartbeatTimeout(ClientClosedConnection):
    def __init__(self, timeout: Union[int, float]):
        self.timeout = timeout  # type: Union[int, float]

    def __str__(self):
        return 'Client closed connection because server did not send any frames in {} seconds'.format(self.timeout)


class InvalidServerFrame(ClientClosedConnection):
    def __init__(self, parse_exc: InvalidParse):
        self.parse_exc = parse_exc  # type: InvalidParse

    def __str__(self):
        return 'Client closed connection because server sent invalid frame data: {}'.format(self.parse_exc)


class UnexpectedFrame(ClientClosedConnection):
    def __init__(self, frame: Frame):
        self.frame = frame  # type: Frame

    def __str__(self):
        return 'Client closed connection because server sent an unexpected frame: {}'.format(self.frame)



from ammoo.wire.class_and_method import cam_long_uint, pretty_format_cam
from ammoo.wire.typing import ReplyCode
from ammoo.wire.frames.method.channel import ChannelCloseParameters

__all__ = (
    'ChannelClosed', 'ServerClosedChannel', 'ClientClosedChannel', 'ClientClosedChannelOK', 'ClientClosedChannelError',
    'EmptyQueue', 'PublishNack', 'ChannelInactive'
)


class ChannelClosed(Exception):
    def __init__(self):
        raise Exception('Do not instantiate me directly')


class ServerClosedChannel(ChannelClosed):
    def __init__(self, parameters: ChannelCloseParameters):
        self.reply_code = parameters.reply_code  # type: ReplyCode
        self.reply_text = parameters.reply_text  # type: str
        self.class_id = parameters.class_id  # type: int
        self.method_id = parameters.method_id  # type: int

    def __str__(self) -> str:
        if self.class_id != 0 and self.method_id != 0:
            cam_text = ' because of method {}'.format(pretty_format_cam(cam_long_uint(self.class_id, self.method_id)))
        else:
            cam_text = ''
        return 'Server closed channel with code {}{}: {}'.format(self.reply_code, cam_text, self.reply_text)


class ClientClosedChannel(ChannelClosed):
    pass


class ClientClosedChannelOK(ClientClosedChannel):
    def __init__(self):
        pass

    def __str__(self):
        return 'Client closed channel without errors'


class ClientClosedChannelError(ClientClosedChannel):
    def __init__(self, exc: Exception):
        self.exc = exc  # type: Exception

    def __str__(self):
        return 'Client closed channel because of an unhandled exception: {!r}'.format(self.exc)


class EmptyQueue(Exception):
    def __init__(self):
        pass

    def __str__(self):
        return 'Queue is empty'


class PublishNack(Exception):
    def __init__(self):
        pass

    def __str__(self):
        return 'Publishing message failed because server replied with a negative acknowledgement'


class ChannelInactive(Exception):
    def __init__(self):
        pass

    def __str__(self):
        return 'Publishing message failed because the channel is not active'



import json
from typing import Optional, Union
import logging

from ammoo.wire.class_and_method import BASIC_RETURN_CAM, BASIC_DELIVER_CAM, BASIC_GET_OK_CAM
from ammoo.exceptions.connection import UnexpectedFrame
from ammoo.wire.frames import BodyFrame
from ammoo.wire.frames import MethodFrame
from ammoo.wire.frames.header import BasicHeaderProperties, BasicHeaderFrame
from ammoo.wire.frames.method.basic import BasicDeliverParameters
from ammoo.wire.frames.method.basic import BasicGetOkParameters
from ammoo.wire.frames.method.basic import BasicReturnParameters
from ammoo.wire.typing import DecodeJson, RoutingKey, ExchangeName, DeliveryTag, ConsumerTag

__all__ = 'Message', 'ReturnMessage', 'ExpectedMessage', 'DeliverMessage', 'GetMessage', 'AbstractMessage'

logger = logging.getLogger(__name__)


class Message:
    """
    Base class for messages. Not instantiated directly::

        Message: body, decode(), json(), exchange_name, routing_key, properties
        -> ReturnMessage: reply_code, reply_text
        -> ExpectedMessage: ack(), reject(), reply(), delivery_tag, redelivered
           -> DeliverMessage: consumer_tag
           -> GetMessage: message_count

    .. attribute:: body

        Message's raw body as a :class:`bytearray` instance

            >>> message.body
            bytearray(b'binary data')

    .. attribute:: exchange_name

        Exchange message was published to

    .. attribute:: routing_key

        Routing key message was published with

    .. attribute:: properties

        Message's :class:`~ammoo.wire.frames.header.BasicHeaderProperties`
    """

    def __init__(
            self, exchange_name: ExchangeName, routing_key: RoutingKey, properties: BasicHeaderProperties,
            body: bytearray, channel_toolbox
    ):
        self.exchange_name = exchange_name
        self.routing_key = routing_key
        self.properties = properties
        self.body = body
        # this is a subset of Channel's attributes. We don't want to give direct access to its internals.
        self._channel_toolbox = channel_toolbox  # a _ChannelToolbox instance

    def decode(self, encoding: Optional[str]=None) -> str:
        """
        Decode body into a :class:`str`. When the encoding argument is not passed, the encoding defaults to the
        message's content-encoding property (if defined), or the channel's default encoding.

        >>> message.decode()
        'text body'
        >>> message.decode('iso-8859-1')
        'sí'

        :param str encoding: Optional: Encoding to use to decode body instead of content-encoding property/channel's \
        default encoding
        :rtype: str
        """
        if encoding is None:
            content_encoding = self.properties.content_encoding
            if content_encoding is None:
                encoding = self._channel_toolbox.default_encoding
            else:
                encoding = content_encoding
        return self.body.decode(encoding)

    def json(self, encoding: Optional[str]=None) -> DecodeJson:
        """
        Decode body as JSON.

        :param str encoding: Optional: Encoding to use to decode body instead of content-encoding property/channel's \
        default encoding
        :rtype: str, bool, int, float, None, dict, list
        """
        return json.loads(self.decode(encoding))


class ReturnMessage(Message):
    """
    Message returned by server as a consequence of using the `mandatory` or `immediate` flags of
    :meth:`Channel.publish`. Subclass of :class:`Message`.

    .. attribute:: reply_code

        :class:`int` code for why message could not be routed to queue/consumed.

    .. attribute:: reply_text

        :class:`str` description of why message was returned.
    """

    __slots__ = 'exchange_name', 'routing_key', 'properties', 'body', '_channel_toolbox', 'reply_code', 'reply_text'
    cam = BASIC_RETURN_CAM

    def __init__(
            self, exchange_name: ExchangeName, routing_key: RoutingKey, properties: BasicHeaderProperties,
            body: bytearray, channel_toolbox, reply_code: int, reply_text: str):
        super().__init__(exchange_name, routing_key, properties, body, channel_toolbox)
        self.reply_code = reply_code
        self.reply_text = reply_text


class ExpectedMessage(Message):
    """
    Base class for :class:`DeliverMessage` and :class:`GetMessage`; not instantiated directly. Subclass of
    :class:`Message`.

    .. attribute:: delivery_tag

        Message's delivery tag, used for acknowledging or rejecting message to server

        .. note:: Using the :meth:`ack` or :meth:`reject` methods of this class instead of :class:`Channel`'s avoids \
        needing to pass the delivery tag explicitly.

    .. attribute: redelivered

        :class:`bool` that is true when the message has been redelivered to queue because it was rejected or \
        dead-lettered before.

    """
    def __init__(
            self, exchange_name: ExchangeName, routing_key: RoutingKey, properties: BasicHeaderProperties,
            body: bytearray, channel_toolbox, delivery_tag: DeliveryTag, redelivered: bool):
        super().__init__(exchange_name, routing_key, properties, body, channel_toolbox)
        self.delivery_tag = delivery_tag
        self.redelivered = redelivered

    async def ack(self) -> None:
        """
        Acknowledge message to server. Calls :meth:`Channel.ack` with the message's delivery tag.
        """
        return await self._channel_toolbox.ack(self.delivery_tag)

    async def reject(self, requeue: bool) -> None:
        """
        Reject message to server. Calls :meth:`Channel.reject` with the message's delivery tag.
        """
        return await self._channel_toolbox.reject(self.delivery_tag, requeue=requeue)

    async def reply(self, body=None, *, correlation_id=None, **kwargs) -> None:
        """
        Publish a reply to a message that has the reply-to property set. If the message has the correlation-id
        property, it's also set on the published message.

        The method accepts the same keyword arguments as :meth:`Channel.publish`.

        .. note:: Direct reply to is a RabbitMQ extension
        """
        # NB: reply_to is the routing key the server should reply to, not the one the client will send
        if not self._channel_toolbox.server_has_capability('direct_reply_to'):
            logger.warn('Server does not support capability direct_reply_to')
        reply_to = self.properties.reply_to
        if reply_to is None:
            raise ValueError('Message has no reply_to property')
        if correlation_id is None:
            correlation_id = self.properties.correlation_id
        return await self._channel_toolbox.publish('', reply_to, body, correlation_id=correlation_id, **kwargs)


class DeliverMessage(ExpectedMessage):
    """
    Message delivered to a :class:`Consumer`. Subclass of :class:`ExpectedMessage`.

    .. attribute:: consumer_tag

        :class:`str` consumer tag parameter of delivered message.
    """

    __slots__ = (
        'exchange_name', 'routing_key', 'properties', 'body', '_channel_toolbox', 'delivery_tag', 'redelivered',
        'consumer_tag'
    )
    cam = BASIC_DELIVER_CAM

    def __init__(
            self, exchange_name: ExchangeName, routing_key: RoutingKey, properties: BasicHeaderProperties,
            body: bytearray, channel_toolbox, delivery_tag: DeliveryTag, redelivered: bool, consumer_tag: ConsumerTag):
        super().__init__(exchange_name, routing_key, properties, body, channel_toolbox, delivery_tag, redelivered)
        self.consumer_tag = consumer_tag


class GetMessage(ExpectedMessage):
    """
    A message from queue returned by calling :meth:`Channel.get`. Subclass of :class:`ExpectedMessage`.

    .. attribute:: message_count

        Number of messages still in queue after getting this message.
    """

    cam = BASIC_GET_OK_CAM
    __slots__ = (
        'exchange_name', 'routing_key', 'properties', 'body', '_channel_toolbox', 'delivery_tag', 'redelivered',
        'message_count'
    )

    def __init__(
            self, exchange_name: ExchangeName, routing_key: RoutingKey, properties: BasicHeaderProperties,
            body: bytearray, channel_toolbox, delivery_tag: DeliveryTag, redelivered: bool, message_count: int):
        super().__init__(exchange_name, routing_key, properties, body, channel_toolbox, delivery_tag, redelivered)
        self.message_count = message_count


# typing to exclude superclasses
AbstractMessage = Union[DeliverMessage, GetMessage, ReturnMessage]


class PartialMessage:
    def __init__(self, channel_toolbox):
        # parameters are basic_get_ok_parameters_pb or basic_deliver_parameters_pb or...
        # basic_return_pb, in which case it doesn't have delivery-tag or some other things
        self._cam = None  # type:  Optional[int]
        self.parameters = None  # type: Union[BasicDeliverParameters, BasicGetOkParameters, BasicReturnParameters]
        self.properties = None  # type: Optional[BasicHeaderProperties]
        self.body = None  # type:  Optional[bytes]
        self._body_offset = None  # type:  Optional[int]
        self._body_size = None  # type:  Optional[int]
        self._channel_toolbox = channel_toolbox

    def _assemble_message(self) -> AbstractMessage:
        """Assembles the method, header and body frames into a message, then resets self ready for the next message
        to be received.
        """
        parameters = self.parameters
        if self._cam == BASIC_DELIVER_CAM:
            message = DeliverMessage(
                parameters.exchange_name, parameters.routing_key, self.properties, self.body, self._channel_toolbox,
                parameters.delivery_tag, parameters.redelivered, parameters.consumer_tag
            )
        elif self._cam == BASIC_GET_OK_CAM:
            message = GetMessage(
                parameters.exchange_name, parameters.routing_key, self.properties, self.body, self._channel_toolbox,
                parameters.delivery_tag, parameters.redelivered, parameters.message_count
            )
        elif self._cam == BASIC_RETURN_CAM:
            message = ReturnMessage(
                parameters.exchange_name, parameters.routing_key, self.properties, self.body, self._channel_toolbox,
                parameters.reply_code, parameters.reply_text
            )
        else:
            raise ValueError(self._cam)
        self._cam = None
        self.parameters = None
        self.properties = None
        self.body = None
        self._body_offset = None
        self._body_size = None
        return message

    def feed_method_frame(self, method_frame: MethodFrame) -> None:
        self._cam = method_frame.cam
        self.parameters = method_frame.parameters
        self.properties = None
        self.body = None
        self._body_offset = None
        self._body_size = None

    def feed_header_frame(self, header_frame: BasicHeaderFrame) -> Optional[AbstractMessage]:
        if self.parameters is None:
            raise UnexpectedFrame(header_frame)
        # XXX: header_frame.class_ not there anymore to check
        #if payload.class_ is not self._cam.class_:
        #    raise UnexpectedFrame(header_frame)
        self.properties = header_frame.properties
        body_size = header_frame.body_size
        self._body_size = body_size
        self._body_offset = 0
        if body_size == 0:
            self.body = bytearray()
            return self._assemble_message()
        return None

    def feed_body_frame(self, body_frame: BodyFrame) -> Optional[AbstractMessage]:
        if self.properties is None:
            raise UnexpectedFrame(body_frame)
        payload = body_frame.payload
        payload_size = body_frame.payload_size
        body_offset = self._body_offset
        body_size = self._body_size
        next_body_offset = body_offset + payload_size
        if next_body_offset > body_size:
            logger.warning('Server sent more body data than header frame specified')
            raise UnexpectedFrame(body_frame)
        if body_offset == 0:
            if payload_size == body_size:
                # first body frame contains all of the body data
                self.body = payload
                return self._assemble_message()
            else:
                self.body = bytearray(body_size)
        self.body[body_offset:next_body_offset] = payload
        if next_body_offset == body_size:
            return self._assemble_message()
        self._body_offset = next_body_offset
        return None



from typing import Set, Union

from ammoo.auth.mechanisms.sasl_plain import SaslPlainAuth
from .abc import Authentication
from .mechanisms.amqplain import AMQPlainAuth
from ammoo.exceptions.connection import ServerAuthMechanismsNotSupportedError


class PasswordAuthentication(Authentication):
    def __init__(self, login, password):
        self.login = login
        self.password = password

    def select_mechanism(self, server_mechanisms: Set[str]) -> Union[AMQPlainAuth, SaslPlainAuth]:
        for auth_mech in (AMQPlainAuth, SaslPlainAuth):
            if auth_mech.name() in server_mechanisms:
                return auth_mech(self.login, self.password)
        raise ServerAuthMechanismsNotSupportedError()



import calendar
from datetime import datetime, timezone
from decimal import Decimal
from io import BytesIO
from typing import Sequence

from ammoo.wire.typing import DeliveryMode
from .exceptions import InvalidBoolean, InvalidDeliveryMode
from .structs import parse_octet_uint, parse_long_int, parse_long_long_uint, pack_octet_uint, pack_long_long_uint


def parse_boolean(bio: BytesIO) -> bool:
    value = parse_octet_uint(bio)
    if value == 1:
        return True
    if value == 0:
        return False
    raise InvalidBoolean(value)


TRUE_DATA = pack_octet_uint(1)
BOOLEAN_FALSE_DATA = pack_octet_uint(0)


def pack_boolean(value: bool) -> bytes:
    if value:
        return TRUE_DATA
    return BOOLEAN_FALSE_DATA


_DECIMAL_10 = Decimal(10)


def parse_decimal(bio: BytesIO) -> Decimal:
    scale = parse_octet_uint(bio)
    integer = parse_long_int(bio)
    value = Decimal(integer)
    if scale > 0:
        value *= _DECIMAL_10 ** -scale
    return value


def parse_timestamp(bio: BytesIO) -> datetime:
    return datetime.utcfromtimestamp(parse_long_long_uint(bio)).replace(tzinfo=timezone.utc)


def pack_timestamp(timestamp: datetime) -> bytes:
    return pack_long_long_uint(calendar.timegm(timestamp.utctimetuple()))


def build_timestamp(bio: BytesIO, timestamp: datetime) -> None:
    bio.write(pack_timestamp(timestamp))


def parse_no_field(bio: BytesIO) -> None:
    return None


DELIVERY_MODE_NON_PERSISTENT = 1
DELIVERY_MODE_PERSISTENT = 2


def parse_delivery_mode(bio: BytesIO) -> DeliveryMode:
    delivery_mode = parse_octet_uint(bio)
    if delivery_mode != DELIVERY_MODE_NON_PERSISTENT and delivery_mode != DELIVERY_MODE_PERSISTENT:
        raise InvalidDeliveryMode(delivery_mode)
    return delivery_mode


def parse_booleans(bio: BytesIO, n: int) -> Sequence[bool]:
    if n > 8:
        raise NotImplementedError()
    byte = parse_octet_uint(bio)
    return tuple(bool(byte & (1 << i)) for i in range(n))


def pack_booleans(*booleans: Sequence[bool]) -> bytes:
    if len(booleans) > 8:
        raise ValueError('Max one byte of booleans is supported')
    byte = 0
    for i, boolean in enumerate(booleans):
        if boolean:
            byte |= 1 << i
    return pack_octet_uint(byte)



from typing import NewType

from ammoo.wire import methods as m
from ammoo.wire.low.structs import parse_long_uint
from . import classes as c


CAM = NewType('CAM', int)

parse_class_and_method = parse_long_uint

TWO_POW16 = 2 ** 16


def cam_long_uint(class_id: int, method_id: int ) -> CAM:
    """Combines a 16-bit class id and a 16-bit method id into a 32-bit integer. This allows to parse the class and
    method directly into a 32-bit int from the reader, and use it as a dictionary key instead of a 2-tuple."""
    return class_id * TWO_POW16 + method_id


CONNECTION_START_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_START)
CONNECTION_START_OK_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_START_OK)
CONNECTION_SECURE_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_SECURE)
CONNECTION_SECURE_OK_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_SECURE_OK)
CONNECTION_TUNE_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_TUNE)
CONNECTION_TUNE_OK_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_TUNE_OK)
CONNECTION_OPEN_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_OPEN)
CONNECTION_OPEN_OK_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_OPEN_OK)
CONNECTION_CLOSE_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_CLOSE)
CONNECTION_CLOSE_OK_CAM = cam_long_uint(c.CLASS_CONNECTION, m.METHOD_CONNECTION_CLOSE_OK)

CHANNEL_OPEN_CAM = cam_long_uint(c.CLASS_CHANNEL, m.METHOD_CHANNEL_OPEN)
CHANNEL_OPEN_OK_CAM = cam_long_uint(c.CLASS_CHANNEL, m.METHOD_CHANNEL_OPEN_OK)
CHANNEL_FLOW_CAM = cam_long_uint(c.CLASS_CHANNEL, m.METHOD_CHANNEL_FLOW)
CHANNEL_FLOW_OK_CAM = cam_long_uint(c.CLASS_CHANNEL, m.METHOD_CHANNEL_FLOW_OK)
CHANNEL_CLOSE_CAM = cam_long_uint(c.CLASS_CHANNEL, m.METHOD_CHANNEL_CLOSE)
CHANNEL_CLOSE_OK_CAM = cam_long_uint(c.CLASS_CHANNEL, m.METHOD_CHANNEL_CLOSE_OK)

EXCHANGE_DECLARE_CAM = cam_long_uint(c.CLASS_EXCHANGE, m.METHOD_EXCHANGE_DECLARE)
EXCHANGE_DECLARE_OK_CAM = cam_long_uint(c.CLASS_EXCHANGE, m.METHOD_EXCHANGE_DECLARE_OK)
EXCHANGE_DELETE_CAM = cam_long_uint(c.CLASS_EXCHANGE, m.METHOD_EXCHANGE_DELETE)
EXCHANGE_DELETE_OK_CAM = cam_long_uint(c.CLASS_EXCHANGE, m.METHOD_EXCHANGE_DELETE_OK)
EXCHANGE_BIND_CAM = cam_long_uint(c.CLASS_EXCHANGE, m.METHOD_EXCHANGE_BIND)
EXCHANGE_BIND_OK_CAM = cam_long_uint(c.CLASS_EXCHANGE, m.METHOD_EXCHANGE_BIND_OK)

QUEUE_DECLARE_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_DECLARE)
QUEUE_DECLARE_OK_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_DECLARE_OK)
QUEUE_BIND_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_BIND)
QUEUE_BIND_OK_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_BIND_OK)
QUEUE_PURGE_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_PURGE)
QUEUE_PURGE_OK_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_PURGE_OK)
QUEUE_DELETE_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_DELETE)
QUEUE_DELETE_OK_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_DELETE_OK)
QUEUE_UNBIND_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_UNBIND)
QUEUE_UNBIND_OK_CAM = cam_long_uint(c.CLASS_QUEUE, m.METHOD_QUEUE_UNBIND_OK)

BASIC_QOS_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_QOS)
BASIC_QOS_OK_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_QOS_OK)
BASIC_CONSUME_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_CONSUME)
BASIC_CONSUME_OK_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_CONSUME_OK)
BASIC_CANCEL_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_CANCEL)
BASIC_CANCEL_OK_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_CANCEL_OK)
BASIC_PUBLISH_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_PUBLISH)
BASIC_RETURN_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_RETURN_)
BASIC_DELIVER_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_DELIVER)
BASIC_GET_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_GET)
BASIC_GET_OK_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_GET_OK)
BASIC_GET_EMPTY_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_GET_EMPTY)
BASIC_ACK_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_ACK)
BASIC_REJECT_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_REJECT)
BASIC_RECOVER_ASYNC_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_RECOVER_ASYNC)
BASIC_RECOVER_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_RECOVER)
BASIC_RECOVER_OK_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_RECOVER_OK)
BASIC_NACK_CAM = cam_long_uint(c.CLASS_BASIC, m.METHOD_BASIC_NACK)

CONFIRM_SELECT_CAM = cam_long_uint(c.CLASS_CONFIRM, m.METHOD_CONFIRM_SELECT)
CONFIRM_SELECT_OK_CAM = cam_long_uint(c.CLASS_CONFIRM, m.METHOD_CONFIRM_SELECT_OK)


_CLASS_NAMES = {
    c.CLASS_CONNECTION: 'connection',
    c.CLASS_CHANNEL: 'channel',
    c.CLASS_EXCHANGE: 'exchange',
    c.CLASS_QUEUE: 'queue',
    c.CLASS_BASIC: 'basic',
    c.CLASS_CONFIRM: 'confirm',
    c.CLASS_TX: 'tx',
}

_METHOD_NAMES = {
    c.CLASS_CONNECTION: {
        m.METHOD_CONNECTION_START: 'start',
        m.METHOD_CONNECTION_START_OK: 'start-ok',
        m.METHOD_CONNECTION_SECURE: 'secure',
        m.METHOD_CONNECTION_SECURE_OK: 'secure-ok',
        m.METHOD_CONNECTION_TUNE: 'tune',
        m.METHOD_CONNECTION_TUNE_OK: 'tune-ok',
        m.METHOD_CONNECTION_OPEN: 'open',
        m.METHOD_CONNECTION_OPEN_OK: 'open-ok',
        m.METHOD_CONNECTION_CLOSE: 'close',
        m.METHOD_CONNECTION_CLOSE_OK: 'close-ok',
    },
    c.CLASS_CHANNEL: {
        m.METHOD_CHANNEL_OPEN: 'open',
        m.METHOD_CHANNEL_OPEN_OK: 'open-ok',
        m.METHOD_CHANNEL_FLOW: 'flow',
        m.METHOD_CHANNEL_FLOW_OK: 'flow-ok',
        m.METHOD_CHANNEL_CLOSE: 'close',
        m.METHOD_CHANNEL_CLOSE_OK: 'close-ok',
    },
    c.CLASS_EXCHANGE: {
        m.METHOD_EXCHANGE_DECLARE: 'declare',
        m.METHOD_EXCHANGE_DECLARE_OK: 'declare-ok',
        m.METHOD_EXCHANGE_DELETE: 'delete',
        m.METHOD_EXCHANGE_DELETE_OK: 'delete-ok',
        m.METHOD_EXCHANGE_BIND: 'bind',
        m.METHOD_EXCHANGE_BIND_OK: 'bind-ok',
    },
    c.CLASS_QUEUE: {
        m.METHOD_QUEUE_DECLARE: 'declare',
        m.METHOD_QUEUE_DECLARE_OK: 'declare-ok',
        m.METHOD_QUEUE_BIND: 'bind',
        m.METHOD_QUEUE_BIND_OK: 'bind-ok',
        m.METHOD_QUEUE_PURGE: 'purge',
        m.METHOD_QUEUE_PURGE_OK: 'purge-ok',
        m.METHOD_QUEUE_DELETE: 'delete',
        m.METHOD_QUEUE_DELETE_OK: 'delete-ok',
        m.METHOD_QUEUE_UNBIND: 'unbind',
        m.METHOD_QUEUE_UNBIND_OK: 'unbind-ok',
    },
    c.CLASS_BASIC: {
        m.METHOD_BASIC_QOS: 'qos',
        m.METHOD_BASIC_QOS_OK: 'qos-ok',
        m.METHOD_BASIC_CONSUME: 'consume',
        m.METHOD_BASIC_CONSUME_OK: 'consume-ok',
        m.METHOD_BASIC_CANCEL: 'cancel',
        m.METHOD_BASIC_CANCEL_OK: 'cancel-ok',
        m.METHOD_BASIC_PUBLISH: 'publish',
        m.METHOD_BASIC_RETURN_: 'return',
        m.METHOD_BASIC_DELIVER: 'deliver',
        m.METHOD_BASIC_GET: 'get',
        m.METHOD_BASIC_GET_OK: 'get-ok',
        m.METHOD_BASIC_GET_EMPTY: 'get-empty',
        m.METHOD_BASIC_ACK: 'ack',
        m.METHOD_BASIC_REJECT: 'reject',
        m.METHOD_BASIC_RECOVER_ASYNC: 'recover-async',
        m.METHOD_BASIC_RECOVER: 'recover',
        m.METHOD_BASIC_RECOVER_OK: 'recover-ok',
        m.METHOD_BASIC_NACK: 'nack',
    },
    c.CLASS_CONFIRM: {
        m.METHOD_CONFIRM_SELECT: 'select',
        m.METHOD_CONFIRM_SELECT_OK: 'select-ok',
    }
}


def pretty_format_cam(cam):
    class_id, method_id = divmod(cam, TWO_POW16)
    return '{}.{}'.format(_CLASS_NAMES[class_id], _METHOD_NAMES[class_id][method_id])



from bisect import bisect_left

MAX_CHANNEL_LIMIT = 2**16 - 1


class _Range:
    __slots__ = 'low', 'high'

    def __init__(self, low, high):
        assert low <= high
        self.low = low
        self.high = high

    def __len__(self):
        return self.high - self.low + 1

    def __lt__(self, other):
        return self.high < other.low

    def __iter__(self):
        return iter(range(self.low, self.high + 1))

    def __repr__(self):
        return '[{},{}]'.format(self.low, self.high)


class NoChannelsAvailable(Exception):
    def __init__(self, max_channel):
        self.max_channel = max_channel

    def __str__(self):
        return 'No channel identifier is available: {} channels in use'.format(self.max_channel)


class AlreadyFree(Exception):
    def __str__(self):
        return 'channel id is already free'


class ChannelIdManager:
    def __init__(self, max_channel):
        if max_channel is None:
            max_channel = MAX_CHANNEL_LIMIT
        elif max_channel < 1 or max_channel > MAX_CHANNEL_LIMIT:
            raise ValueError(max_channel)
        self._max_channel = max_channel
        self._ranges = [_Range(1, max_channel)]

    @property
    def channel_max(self):
        return self._max_channel

    def acquire(self):
        try:
            rng = self._ranges[0]
        except IndexError:
            raise NoChannelsAvailable(self._max_channel)
        channel_id = rng.low
        if len(rng) == 1:
            del self._ranges[0]
        else:
            rng.low += 1
        return channel_id

    def release(self, channel_id):
        rng = _Range(channel_id, channel_id)
        index = bisect_left(self._ranges, rng)
        try:
            right_neighbor = self._ranges[index]
        except IndexError:
            # no neighbors to the right
            pass
        else:
            if right_neighbor.low <= channel_id:
                raise AlreadyFree()
            if right_neighbor.low == channel_id + 1:
                # merge with it
                right_neighbor.low = rng.low
                rng = right_neighbor
        try:
            left_neighbor = self._ranges[index - 1]
        except IndexError:
            # no neighbors to the left
            pass
        else:
            if left_neighbor.high == channel_id - 1:
                # merge with it
                left_neighbor.high = rng.high
                if len(rng) > 1:
                    # also merged with right - need to remove it
                    del self._ranges[index]
                rng = left_neighbor
        if len(rng) == 1:
            # did not merge with left nor right
            self._ranges.insert(index, rng)

    def __len__(self):
        return sum(len(rng) for rng in self._ranges)

    def __iter__(self):
        for rng in self._ranges:
            yield from rng



from asyncio import StreamReader
from io import BytesIO

from ammoo.wire.class_and_method import parse_class_and_method, pretty_format_cam, cam_long_uint
from ammoo.wire import classes as c
from ammoo.wire.low.exceptions import ExcessiveFramePayload, UnsupportedMethod
from .basic import BASIC_METHOD_PARAMETER_PARSERS
from .channel import CHANNEL_METHOD_PARAMETER_PARSERS
from .confirm import CONFIRM_METHOD_PARAMETER_PARSERS
from .exchange import EXCHANGE_METHOD_PARAMETER_PARSERS
from .queue import QUEUE_METHOD_PARAMETER_PARSERS
from .connection import CONNECTION_METHOD_PARAMETER_PARSERS


METHOD_PARAMETER_PARSERS = {
    c.CLASS_CONNECTION: CONNECTION_METHOD_PARAMETER_PARSERS,
    c.CLASS_CHANNEL: CHANNEL_METHOD_PARAMETER_PARSERS,
    c.CLASS_EXCHANGE: EXCHANGE_METHOD_PARAMETER_PARSERS,
    c.CLASS_QUEUE: QUEUE_METHOD_PARAMETER_PARSERS,
    c.CLASS_BASIC: BASIC_METHOD_PARAMETER_PARSERS,
    c.CLASS_CONFIRM: CONFIRM_METHOD_PARAMETER_PARSERS,
    # Not used by client: c.CLASS_TX: TX_METHOD_PARAMETER_PARSERS,
}
CAM_PARAMETER_PARSERS = {}
for class_id, method_parsers in METHOD_PARAMETER_PARSERS.items():
    for method_id, parse_parameters in method_parsers.items():
        CAM_PARAMETER_PARSERS[cam_long_uint(class_id, method_id)] = parse_parameters


FRAME_TYPE_METHOD = 1


class MethodFrame:
    __slots__ = 'channel_id', 'cam', 'parameters'
    frame_type = FRAME_TYPE_METHOD

    def __init__(self, channel_id: int, cam: int, parameters):  # XXX: typing missing for parameters
        self.channel_id = channel_id
        self.cam = cam
        self.parameters = parameters

    def __str__(self):
        return 'Method {} frame channel {}, parameters: {}'.format(
            pretty_format_cam(self.cam), self.channel_id, self.parameters
        )


async def parse_method_frame(
        reader: StreamReader, payload_size: int, channel_id: int, encoding: str, rabbitmq: bool) -> MethodFrame:
    bio = BytesIO(await reader.readexactly(payload_size))
    cam = parse_class_and_method(bio)
    try:
        parse_parameters = CAM_PARAMETER_PARSERS[cam]
    except KeyError:
        raise UnsupportedMethod(cam)
    if parse_parameters is None:
        parameters = None
    else:
        parameters = parse_parameters(bio, encoding, rabbitmq)
    if bio.tell() != payload_size:
        raise ExcessiveFramePayload(payload_size, bio.tell())
    return MethodFrame(channel_id, cam, parameters)



from io import BytesIO

from ammoo.wire import methods as m
from ammoo.wire.low.misc import parse_boolean
from ammoo.wire.typing import NoWait


def parse_confirm_select_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> NoWait:
    return parse_boolean(bio)


CONFIRM_METHOD_PARAMETER_PARSERS = {
    m.METHOD_CONFIRM_SELECT: parse_confirm_select_parameters,
    m.METHOD_CONFIRM_SELECT_OK: None,
}



from ammoo.wire.constants import FRAME_END


class InvalidParse(Exception):
    """Superclass for all parsing errors"""


class IncompleteFrame(InvalidParse):
    def __init__(self):
        pass


class InvalidFloat(InvalidParse):
    def __init__(self, data):
        self.data = data

    def __str__(self):
        return 'Could not parse a float from {!r}'.format(self.data)


class InvalidDouble(InvalidParse):
    def __init__(self, data):
        self.data = data

    def __str__(self):
        return 'Could not parse a double from {!r}'.format(self.data)


class ShortstrDecodingError(InvalidParse):
    def __init__(self, data, encoding):
        self.data = data
        self.encoding = encoding

    def __str__(self):
        return 'Could not decode shortstr data as {}: {!r}'.format(self.encoding, self.data)


class LongstrDecodingError(InvalidParse):
    def __init__(self, data, encoding):
        self.data = data
        self.encoding = encoding

    def __str__(self):
        return 'Could not decode longstr data as {}: {!r}'.format(self.encoding, self.data)


class InvalidFieldValueType(InvalidParse):
    """Field tables and arrays have a one byte character to specify what type the subsequent value has. Except there
    was an value we don't support."""

    def __init__(self, type_byte):
        self.type_byte = type_byte

    def __str__(self):
        return 'Invalid field value type byte {!r}'.format(self.type_byte)


class InvalidBoolean(InvalidParse):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return 'Could not parse boolean (0 or 1) from value {}'.format(self.value)


class InvalidBooleanTable(InvalidParse):
    def __init__(self, field_type_byte):
        self.field_type_byte = field_type_byte

    def __str__(self):
        return 'Expected a field table with boolean values, not type byte {!r}'.format(self.field_type_byte)


class InvalidDeliveryMode(InvalidParse):
    """Allowed delivery modes are 1 for non-persistent and 2 for persistent, nothing else"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return 'Delivery mode was not 1 or 2: {}'.format(self.value)


class InvalidWeight(InvalidParse):
    """Spec says weight is always zero. Except in this case."""

    def __init__(self, weight):
        self.weight = weight

    def __str__(self):
        return 'Weight is not zero: {}'.format(self.weight)


class InvalidFrameEnd(InvalidParse):
    def __init__(self, data):
        self.data = data

    def __str__(self):
        return 'Could not parse frame end {!r} from data {!r}'.format(FRAME_END, self.data)


class ExcessiveFramePayload(InvalidParse):
    def __init__(self, payload_size: int, read_size: int):
        self.payload_size = payload_size
        self.read_size = read_size

    def __str__(self) -> str:
        return 'Frame has {} bytes of payload, but only {} of it was read'.format(self.payload_size, self.read_size)


class InvalidFrameType(InvalidParse):
    """Raised when the frame type byte in the frame header is unknown"""

    def __init__(self, frame_type: int):
        self.frame_type = frame_type

    def __str__(self):
        return 'Invalid frame type {}'.format(self.frame_type)


class UnsupportedHeaderFrameClass(InvalidParse):
    """AMQP 0.9.1 only specifies Header frames for the basic class. This means we got one for another class."""

    def __init__(self, class_id: int):
        self.class_id = class_id

    def __str__(self) -> str:
        return 'Header frames are only supported for class basic, not {}'.format(self.class_id)


class TooManyHeaderProperties(InvalidParse):
    """The sequence of header properties is fixed for each AMQP class. Server sent flags indicating there's more than
    there should be."""

    def __init__(self, class_id: int):
        self.class_id = class_id

    def __str__(self) -> str:
        return 'Header frame with more properties than class {} specifies'.format(self.class_id)


class UnsupportedMethod(InvalidParse):
    """Server sent a class+method pair we don't understand"""

    def __init__(self, cam: int):
        self.cam = cam

    def __str__(self):
        class_id, method_id = divmod(self.cam, 2**16)
        return 'Method frame with unsupported method {} for class {}'.format(method_id, class_id)



from collections import OrderedDict, deque
from typing import NewType, TypeVar, List, Dict, Union, MutableSequence, _geqv

VirtualHost = NewType('VirtualHost', str)
ConsumerTag = NewType('ConsumerTag', str)
QueueName = NewType('QueueName', str)
ExchangeName = NewType('ExchangeName', str)
ExchangeType = NewType('ExchangeType', str)
RoutingKey = NewType('RoutingKey', str)

ReplyCode = NewType('ReplyCode', int)
DeliveryTag = NewType('DeliveryTag', int)
MessageCount = NewType('MessageCount', int)
ConsumerCount = NewType('ConsumerCount', int)
DeliveryMode = NewType('DeliveryMode', int)

Requeue = NewType('Requeue', bool)
FlowActive = NewType('Active', bool)
NoWait = NewType('NoWait', bool)

# could not use Union because of recursive typing with List and Dict.
# when decoding json, the dict key is always an str, but for encoding, the type can be any json value except list/dict
DecodeJson = TypeVar('DecodeJson', None, int, float, str, List['DecodeJson'], Dict[str, 'DecodeJson'])
JsonKey = Union[None, int, float, str]
EncodeJson = TypeVar('EncodeJson', None, int, float, str, List['DecodeJson'], Dict[JsonKey, 'DecodeJson'])

T = TypeVar('T')  # Any type.
KT = TypeVar('KT')  # Key type.
VT = TypeVar('VT')  # Value type.


class Deque(deque, MutableSequence[T], extra=deque):
    def __new__(cls, *args, **kwds):
        if _geqv(cls, Deque):
            raise TypeError("Type Deque cannot be instantiated; "
                            "use deque() instead")
        return deque.__new__(cls, *args, **kwds)


class OrderedDictT(OrderedDict, Dict[KT, VT], extra=OrderedDict):
    def __new__(cls, *args, **kwds):
        if _geqv(cls, OrderedDict):
            raise TypeError("Type OrderedDictT cannot be instantiated; "
                            "use OrderedDict() instead")
        return OrderedDict.__new__(cls, *args, **kwds)



import asyncio
import logging
import platform
import time
from collections import deque
from functools import partial
from struct import pack
from traceback import format_exc
from typing import Dict, Union
from typing import MutableSequence
from typing import Optional

from ammoo.auth.abc import Authentication, AuthenticationMechanism
from ammoo.channel import Channel
from ammoo.channel_id_manager import ChannelIdManager, AlreadyFree
from ammoo.exceptions import connection as connection_exceptions
from ammoo.exceptions.connection import UnexpectedFrame, ConnectionClosed, ServerAuthMechanismsNotSupportedError, \
    ServerLocalesNotSupportedError, ClientClosedConnectionError
from ammoo.validatation import validate_frame_max, validate_shortstr
from ammoo.version import __version__
from ammoo.wire import constants
from ammoo.wire.class_and_method import CONNECTION_CLOSE_OK_CAM, CONNECTION_CLOSE_CAM, CONNECTION_START_CAM, \
    CONNECTION_SECURE_CAM, CONNECTION_TUNE_CAM, CONNECTION_OPEN_OK_CAM, CONNECTION_START_OK_CAM, pretty_format_cam, \
    CONNECTION_OPEN_CAM, CONNECTION_SECURE_OK_CAM, CONNECTION_TUNE_OK_CAM
from ammoo.wire.frames import FRAME_TYPE_HEARTBEAT
from ammoo.wire.frames import FRAME_TYPE_METHOD
from ammoo.wire.frames import build_frame
from ammoo.wire.frames import parse_frame
from ammoo.wire.frames.method import MethodFrame
from ammoo.wire.frames.method.connection import pack_connection_start_ok_parameters, pack_connection_open_parameters, \
    pack_connection_tune_ok_parameters, CLEAN_CONNECTION_CLOSE_PARAMETERS
from ammoo.wire.low.exceptions import InvalidParse
from ammoo.wire.low.peer_properties import PeerProperties
from ammoo.wire.low.strings import pack_binary_longstr
from ammoo.wire.typing import VirtualHost

__all__ = 'Connection',


logger = logging.getLogger(__name__)


class Connection(asyncio.StreamReaderProtocol):
    amqp_encoding = 'utf-8'

    client_properties = PeerProperties(
        host=None,
        product='ammoo',
        version=__version__,
        # use just platform.system() and not platform.platform() to not reveal exact OS details to peer
        platform='Python {} on {}'.format(platform.python_version(), platform.system()),
        copyright='Mathias Frojdman 2016',
        information=None,  # TODO: add URL here
        capabilities={
            'authentication_failure_close': True,  # https://www.rabbitmq.com/auth-notification.html
            'consumer_cancel_notify': True,  # https://www.rabbitmq.com/consumer-cancel.html,
        },
        others={}
    )

    def __init__(
            self,
            virtualhost: VirtualHost,
            loop: asyncio.AbstractEventLoop,
            heartbeat_interval: Optional[int],
            auth: Union[Authentication, AuthenticationMechanism],
            frame_max: Optional[int],
            body_encoding: str='utf-8'
    ):
        validate_shortstr(virtualhost, 'virtualhost')
        if not isinstance(loop, asyncio.AbstractEventLoop):
            raise TypeError('loop is not an asyncio event loop: {!r}'.format(loop))
        if heartbeat_interval is not None:
            if not isinstance(heartbeat_interval, int):
                raise TypeError()
            if heartbeat_interval < 0:
                raise ValueError('heartbeat_interval must be >= 0: Not {}'.format(heartbeat_interval))
        if not isinstance(auth, (Authentication, AuthenticationMechanism)):
            raise TypeError('auth is not a Authentication or AuthenticationMechanism instance: {!r}'.format(auth))
        if frame_max is not None:
            validate_frame_max(frame_max)
        self._reader = None  # type: Optional[asyncio.StreamReader] # set in _client_connected()
        self._writer = None  # type: Optional[asyncio.StreamWriter] # set in _client_connected()
        self._transport = None  # type: Optional[asyncio.Transport]  # set in connection_made()
        self._dispatcher = None  # type: Optional[Task]  # set in __aenter__()
        self._entered = False  #  type: bool  # set to true in __aenter__()
        self._virtualhost = virtualhost  #  type: VirtualHost
        # the heartbeat interval suggested by the user of this library
        self._client_heartbeat_interval = heartbeat_interval  # type: Optional[int]
        # heartbeat interval negotiated with the server. May be different from self._client_heartbeat_interval
        # set in _tune()
        self._heartbeat_interval = None  #  type: Optional[int]
        # timestamp of last sent frame. Used to determine when the next heartbeat needs to be sent. Set in _send_frame
        self._last_send_timestamp = None  # type: Optional[float]
        # timestamp of last received frame. Used to determine when heartbeat timeout fires
        self._last_receive_timestamp = None  # set in _dispatch_frames()
        # task sending heartbeats to broker, set in tune()
        self._heartbeat_sender = None  # type: Optional[asyncio.Task]
        # task using _last_receive_timestamp to check broker sends frames regularly, set in tune()
        self._heartbeat_monitor = None  # type: Optional[asyncio.Task]
        # frame max suggested by the user of this library
        self._client_frame_max = frame_max  # type: Optional[int]
        # frame max negotiated with broker, set in _tune()
        self._frame_max = None  # type: Optional[int]
        self._loop = loop
        # Authentication used for login with connection.start(-ok) and connection.secure(-ok) methods
        self.auth = auth
        # Keeps track of unused channel ids
        self._channel_id_manager = None # set in _tune()
        # Exception is set when connection is closing/closed.
        self._closing_exc = None
        self._channels = {}  # channel id -> Channel
        self._server_properties = None  # set in _enter_start()
        # Queues for method frame waiters. Library sends a class.method frame, adds a new future to the queue and waits
        # until it is resolved by the dispatcher receiving a class.method-ok frame and setting it as a result.
        # Having a queue of futures means it's possible to wait for multiple method frames of the same type without
        # locks. They're resolved in the same order they were sent out to the broker. Not really used in Connection,
        # but Channel uses plenty of it (send many basic.get methods before receiving basic.get-ok/basic.empty,
        # declare multiple queues etc).
        secure_or_tune_deque = deque()  # type: Dict[int, MutableSequence[asyncio.Future]]
        self._cam_waiters = {
            CONNECTION_CLOSE_OK_CAM: deque(),
            CONNECTION_START_CAM: deque(),
            CONNECTION_SECURE_CAM: secure_or_tune_deque,
            CONNECTION_TUNE_CAM: secure_or_tune_deque,
            CONNECTION_OPEN_OK_CAM: deque(),
        }
        # RabbitMQ and Qpid has their own spec, different from vanilla AMQP 0.9.1 for table/array field types.
        # Only some field types are supported in RabbitMQ/Qpid, and some have a different type byte.
        self.rabbitmq = None  # type: Optional[bool]
        # Used by _drain_writer. Not in the shower.
        self._drain_lock = asyncio.Lock(loop=loop)
        self.body_encoding = body_encoding
        asyncio.StreamReaderProtocol.__init__(
            self, asyncio.StreamReader(loop=self._loop), self._client_connected, loop=self._loop)

    @property
    def heartbeat_interval(self) -> int:
        heartbeat_interval = self._heartbeat_interval
        if heartbeat_interval is None:
            raise ValueError('heartbeat_interval not yet set for this connection')
        return heartbeat_interval

    @property
    def frame_max(self) -> int:
        frame_max = self._frame_max
        if frame_max is None:
            raise ValueError('frame_max not yet set for this connection')
        return frame_max

    def _client_connected(self, reader, writer):
        self._reader = reader
        self._writer = writer

    def connection_made(self, transport):
        self._transport = transport
        super().connection_made(transport)

    def connection_lost(self, exc):
        logger.debug('Connection lost: %r', exc)
        if self._dispatcher is not None:
            self._dispatcher.cancel()
            # it will be set to None in exit()
        # exc is None when client closes the connection -> do not call _close again
        if exc is not None:
            self._close(connection_exceptions.ConnectionLost(exc))
        super().connection_lost(exc)

    @property
    def entered(self) -> bool:
        return self._entered

    @property
    def closing(self) -> bool:
        return self._closing_exc is not None

    def _close(self, close_exc: ConnectionClosed):
        """Only call on errors when connection needs to be abruptly closed without notifying the server"""
        logger.debug('_close because %s', close_exc)
        if self.closing:
            logger.debug('Not setting self._closing_exc because it\'s already %s', self._closing_exc)
        else:
            logger.debug('self._closing_exc is None, setting it to %s', close_exc)
            self._closing_exc = close_exc
        if self._transport is not None:
            logger.debug('Closing transport in _close')
            self._transport.close()
            self._transport = None
        logger.debug('Cancelling subscribers because of close exception %s', close_exc)
        for futures in self._cam_waiters.values():
            for future in futures:
                future.set_exception(self._closing_exc)
            futures.clear()
        for channel in list(self._channels.values()):
            channel.release(self._closing_exc)
        assert not self._channels

    def _dispatch_frame(self, frame):
        logger.debug('Received frame: %s', frame)
        self._last_receive_timestamp = time.monotonic()
        if self.closing:
            if frame.frame_type != FRAME_TYPE_METHOD:
                logger.debug('Closing so ignoring frame of type %s', frame.frame_type)
                return
            elif frame.cam not in (CONNECTION_CLOSE_CAM, CONNECTION_CLOSE_OK_CAM):
                logger.debug('Closing so ignoring method %s frame', pretty_format_cam(frame.cam))
                return
        try:
            frame_type = frame.frame_type
            if frame_type == FRAME_TYPE_HEARTBEAT:
                if self._heartbeat_interval is None or self._heartbeat_interval == 0:
                    logger.warning('Server sent heartbeat frame though heartbeats are turned off')
                    raise UnexpectedFrame(frame)
            elif frame.channel_id > 0:
                channel_id = frame.channel_id
                try:
                    channel = self._channels[channel_id]
                except KeyError as exc:
                    logger.warning('Got frame for channel id %s, but it is not open', channel_id)
                    raise UnexpectedFrame(frame) from exc
                else:
                    channel.dispatch_frame(frame)
            elif frame_type == FRAME_TYPE_METHOD:
                cam = frame.cam
                if cam == CONNECTION_CLOSE_CAM:
                    self._on_connection_close(frame)
                elif cam in self._cam_waiters:
                    try:
                        future = self._cam_waiters[cam].popleft()
                    except IndexError:
                        raise UnexpectedFrame(frame)
                    if not future.done():
                        future.set_result(frame)
                else:
                    raise UnexpectedFrame(frame)
            else:
                raise UnexpectedFrame(frame)
        except UnexpectedFrame:
            # TODO: send a 503 or whatnot connection.close to server? Need to wait for connection.close-ok then?
            raise

    async def _dispatch_frames(self):
        """Reads frames from self._reader and dispatches them to handlers"""
        while True:
            try:
                frame = await parse_frame(self._reader, self.amqp_encoding, self.rabbitmq)
            except asyncio.CancelledError:
                raise
            except InvalidParse as exc:
                logger.error('Could not parse server data: %s', exc)
                return self._close(connection_exceptions.InvalidServerFrame(exc))
            except Exception as exc:
                logger.error('Unhandled exception occurred in parsing frame: %s', format_exc())
                return self._close(connection_exceptions.ClientClosedConnectionError(exc))
            try:
                self._dispatch_frame(frame)
            except asyncio.CancelledError:
                raise
            except UnexpectedFrame as exc:
                return self._close(exc)
            except Exception as exc:
                logger.error('Unhandled exception occurred in frame dispatch: %s', format_exc())
                return self._close(connection_exceptions.ClientClosedConnectionError(exc))

    def _on_connection_close(self, frame: MethodFrame) -> None:
        logger.debug('in _on_connection_close')
        if not self.closing:
            logger.debug('Setting _closing_reason to ConnectionClosed by server')
            self._closing_exc = connection_exceptions.ServerClosedConnection(frame.parameters)
        else:
            logger.debug('self.closing is already true, not setting _closing_exc')
        self._send_method_frame(0, CONNECTION_CLOSE_OK_CAM, b'')

    async def __aenter__(self) -> 'Connection':
        if self.entered:
            raise ValueError('Connection already entered')
        self._entered = True
        if self.closing:
            raise self._closing_exc
        self._dispatcher = self._loop.create_task(self._dispatch_frames())
        # run remainder of function in try and do cleanup similar to __aexit__ if it fails
        try:
            self._writer.write(constants.PROTOCOL_HEADER)
            tune_frame = await self._enter_start()
            self._tune(tune_frame)
            await self._connection_open()
            self._entered = True
            logger.debug('enter done!')
        except ConnectionClosed as exc:
            logger.error(format_exc())
            if not self.closing:
                self._closing_exc = exc
            await self._cleanup()
            raise
        except Exception as exc:
            logger.error(format_exc())
            if not self.closing:
                self._closing_exc = connection_exceptions.ClientClosedConnectionError(exc)
            await self._cleanup()
            raise
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self.entered:
            raise ValueError('Connection not entered')
        try:
            if not self.closing:
                self._send_method_frame(0, CONNECTION_CLOSE_CAM, CLEAN_CONNECTION_CLOSE_PARAMETERS)
                self._closing_exc = connection_exceptions.ClientClosedConnectionOK()
                await self._wait_for_cam_frame(CONNECTION_CLOSE_OK_CAM)
        finally:
            await self._cleanup()

    async def _wait_for_cam_frame(self, cam: int) -> MethodFrame:
        future = self._loop.create_future()
        self._cam_waiters[cam].append(future)
        log_debug = logger.isEnabledFor(logging.DEBUG)
        if log_debug:
            logger.debug('Waiting for %s', pretty_format_cam(cam))
        method_frame = await future
        if log_debug:
            logger.debug('Received %s (same key as for %s)', pretty_format_cam(method_frame.cam), pretty_format_cam(cam))
        return method_frame

    async def _cleanup(self) -> None:
        """Called always from exit(). Called from enter() when an exception occured"""
        cancellees = set()
        dispatcher = self._dispatcher
        if dispatcher is not None:
            logger.debug('Cancelling dispatcher')
            if not dispatcher.done():
                dispatcher.cancel()
            self._dispatcher = None
            cancellees.add(dispatcher)
        heartbeat_sender = self._heartbeat_sender
        if heartbeat_sender is not None:
            logger.debug('Cancelling heartbeat sender')
            if not heartbeat_sender.done():
                heartbeat_sender.cancel()
            self._heartbeat_sender = None
            cancellees.add(heartbeat_sender)
        heartbeat_monitor = self._heartbeat_monitor
        if heartbeat_monitor is not None:
            logger.debug('Cancelling heartbeat monitor')
            if not heartbeat_monitor.done():
                heartbeat_monitor.cancel()
            self._heartbeat_monitor = None
            cancellees.add(heartbeat_sender)
        for cancellee in cancellees:
            try:
                await cancellee
            except asyncio.CancelledError:
                pass
            except Exception as exc:
                logger.error('Not a CancelledError: %r', exc)
                raise
        if self._transport is not None:
            logger.debug('Closing transport in _cleanup')
            self._transport.close()
            self._transport = None
        return

    async def _enter_start(self) -> MethodFrame:
        """Waits for connection.start, saves server properties to self, sends uses self.auth to authenticate with
        connection.start-ok. It then replies to connection.secure methods with connection.secure-ok until the broker
        sends a connection.tune methods indicating authentication is done.
        :return: The connection.tune method frame sent by the broker
        """
        start_method_frame = await self._wait_for_cam_frame(CONNECTION_START_CAM)
        server_properties = start_method_frame.parameters.server_properties
        self._server_properties = server_properties
        self.rabbitmq = server_properties.product == 'RabbitMQ'  # TODO: Figure out what product string QPid uses
        if 'en_US' not in start_method_frame.parameters.locales:
            exc = ServerLocalesNotSupportedError(start_method_frame.parameters.locales)
            self._close(exc)
            raise exc
        if isinstance(self.auth, Authentication):
            try:
                auth_mech = self.auth.select_mechanism(start_method_frame.parameters.mechanisms)
            except ServerAuthMechanismsNotSupportedError as exc:
                self._close(exc)
                raise
        else:
            auth_mech = self.auth
            if auth_mech.name() not in start_method_frame.parameters.mechanisms:
                exc = ServerAuthMechanismsNotSupportedError()
                self._close(exc)
                raise exc
        start_ok_response = auth_mech.build_start_ok_response(self.amqp_encoding, self.rabbitmq)  # type: bytes
        start_ok_params = pack_connection_start_ok_parameters(
            self.client_properties,
            auth_mech.name(),
            start_ok_response,
            'en_US',
            self.amqp_encoding
        )
        self._send_method_frame(0, CONNECTION_START_OK_CAM, start_ok_params)

        while True:
            # if our auth in connection.start-ok's response parameter wasn't any good, this is where the server will
            # close the connection
            frame = await self._wait_for_cam_frame(CONNECTION_TUNE_CAM)  # same deque() also for connection.secure
            if frame.cam == CONNECTION_SECURE_CAM:
                try:
                    start_ok_response = await auth_mech.challenge(frame.parameters)
                except ConnectionClosed as exc:
                    # most likely an AuthenticationError
                    self._close(ClientClosedConnectionError(exc))
                    raise
                except Exception as exc:
                    self._close(ClientClosedConnectionError(exc))
                    raise
                if not isinstance(start_ok_response, bytes):
                    exc = ValueError('{!r}.challenge did not return bytes: {!r}'.format(
                        auth_mech.__class__, start_ok_response
                    ))
                    self._close(ClientClosedConnectionError(exc))
                    raise exc
                self._send_method_frame(0, CONNECTION_SECURE_OK_CAM, pack_binary_longstr(start_ok_response))
            else:  # connection.tune
                return frame

    def _tune(self, tune_frame: MethodFrame) -> None:
        """Takes the connection settings suggested by the broker and replies with a connection.tune-ok method.
        Starts tasks for sending heartbeats and monitoring for received ones, if heartbeat_interval > 0.
        """
        tune_params = tune_frame.parameters
        if tune_params.channel_max == 0:
            channel_max = None  # zero means no max
        else:
            channel_max = tune_params.channel_max
        self._channel_id_manager = ChannelIdManager(channel_max)

        if tune_params.frame_max < self._client_frame_max:
            self._frame_max = tune_params.frame_max
        else:
            self._frame_max = self._client_frame_max

        if self._client_heartbeat_interval is None:
            self._heartbeat_interval = tune_params.heartbeat
        elif tune_params.heartbeat == 0:
            self._heartbeat_interval = self._client_heartbeat_interval
        else:
            self._heartbeat_interval = min(self._client_heartbeat_interval, tune_params.heartbeat)
        logger.debug('Setting heartbeat interval to %r', self._heartbeat_interval)

        tune_ok_params = pack_connection_tune_ok_parameters(
            tune_params.channel_max, self._frame_max, self._heartbeat_interval)
        self._send_method_frame(0, CONNECTION_TUNE_OK_CAM, tune_ok_params)
        if self._heartbeat_interval > 0:
            self._heartbeat_sender = self._loop.create_task(self._send_heartbeats())
            self._heartbeat_monitor = self._loop.create_task(self._monitor_heartbeats(self._heartbeat_interval))

    async def _connection_open(self) -> None:
        """Sends a connection.open method, then waits for connection.open-ok"""
        connection_open_params = pack_connection_open_parameters(self._virtualhost, [], False, self.amqp_encoding)
        self._send_method_frame(0, CONNECTION_OPEN_CAM, connection_open_params)
        await self._wait_for_cam_frame(CONNECTION_OPEN_OK_CAM)
        return

    async def _monitor_heartbeats(self, heartbeat_interval: int, *, heartbeat_window: int=2) -> None:
        """
        Closes the connection if broker doesn't send frames (heartbeat or not) in heartbeat_interval *
        heartbeat_multiplier seconds.

        :param heartbeat_interval: Interval in seconds client expects broker to send heartbeats
        :param heartbeat_window: Number of heartbeats allowed to be missed before closing the connection
        """
        heartbeat_timeout = heartbeat_interval * heartbeat_window
        while True:
            delay = self._last_receive_timestamp + heartbeat_timeout - time.monotonic()
            if delay > 0.0:
                await asyncio.sleep(delay, loop=self._loop)
            else:
                break
        if self.closing:
            return
        self._close(connection_exceptions.HeartbeatTimeout(heartbeat_timeout))

    async def _send_heartbeats(self) -> None:
        """
        Coroutine to be run as task started when the heartbeat interval has been negotiated with the broker with
        connection.tune and connection.tune-ok. Sends heartbeat frames every heartbeat_interval seconds, if no other
        frames have been sent in the meanwhile. It trusts they set self._last_send_timestamp, so it can skip sending
        a heartbeat in vain.
        """
        # NB: must not be called before first frame is sent
        while True:
            # this loop takes care of not sending heartbeat frames if other frames have been sent
            while True:
                delay = self._last_send_timestamp + self._heartbeat_interval - time.monotonic()
                if delay > 0.0:
                    await asyncio.sleep(delay, loop=self._loop)
                else:
                    break
            if self.closing:
                break
            else:
                logger.debug('Sending heartbeat')
                self._send_frame(FRAME_TYPE_HEARTBEAT, 0, b'')
        return

    def _send_method_frame(self, channel_id: int, cam: int, parameters: bytes) -> None:
        if self.closing and cam != CONNECTION_CLOSE_OK_CAM:
            raise ValueError('Refusing to send frame when closing')
        payload = pack('!I', cam) + parameters
        build_frame(self._writer, FRAME_TYPE_METHOD, channel_id, payload)
        self._last_send_timestamp = time.monotonic()

    def _send_frame(self, frame_type: int, channel_id: int, payload: bytes) -> None:
        if self.closing:
            raise ValueError('Refusing to send frame when closing')
        build_frame(self._writer, frame_type, channel_id, payload)
        self._last_send_timestamp = time.monotonic()

    async def _drain_writer(self) -> None:
        # need to wrap StreamWriter.drain in a lock to avoid AssertionError in FlowControlMixin._drain_helper
        # https://groups.google.com/forum/#!topic/python-tulip/JA0-FC_pliA
        async with self._drain_lock:
            await self._writer.drain()
        return

    def server_has_capability(self, capability: str) -> bool:
        return self._server_properties.capabilities.get(capability, False)

    def _on_channel_exit(self, channel_id: int) -> None:
        # _on_channel_exit will be called twice from Channel.release for open channels when connection is closed:
        # once by Channel.__aexit__ and once by Connection._close. That's why we need to check if channel_id is still
        # in self._channels and the channel id manager
        if channel_id in self._channels:
            del self._channels[channel_id]
        try:
            self._channel_id_manager.release(channel_id)
        except AlreadyFree:
            pass

    def channel(self, *, prefetch_size: Optional[int]=None, prefetch_count: Optional[int]=None) -> Channel:
        if not self.entered:
            raise ValueError()
        if self.closing:
            raise self._closing_exc
        channel_id = self._channel_id_manager.acquire()
        channel = Channel(
            channel_id,
            self._send_frame,
            self._send_method_frame,
            self._loop,
            partial(self._on_channel_exit, channel_id),
            self._frame_max,
            self.server_has_capability,
            self._drain_writer,
            prefetch_size=prefetch_size,
            prefetch_count=prefetch_count,
            body_encoding=self.body_encoding,
            amqp_encoding=self.amqp_encoding,
            rabbitmq=self.rabbitmq
        )
        self._channels[channel_id] = channel
        return channel



from io import StringIO
from typing import Optional

from ammoo.auth.abc import AuthenticationMechanism


class SaslPlainAuth(AuthenticationMechanism):
    """The SASL PLAIN authentication mechanism according to RFC 4616"""
    __slots__ = 'authcid', 'passwd', 'authzid'

    @classmethod
    def name(self):
        return 'PLAIN'

    def __init__(self, authcid: str, passwd: str, authzid: Optional[str]=None):
        """
        :param authcid: Authentication identity (username)
        :param passwd: Password
        :param authzid: Optional: Authorization identity
        """
        if not isinstance(authcid, str):
            raise TypeError('authcid must be a str object')
        if '\x00' in authcid:
            raise ValueError('NUL is not allowed')
        if not isinstance(passwd, str):
            raise TypeError('passwd must be a str object')
        if authzid is not None and not isinstance(authzid, str):
            raise TypeError('authzid must be a str object')
        self.authcid = authcid
        self.passwd = passwd
        self.authzid = authzid

    def build_start_ok_response(self, encoding, rabbitmq):
        if encoding != 'utf-8':
            raise AuthFail('{} only supports utf-8 encoding'.format(self.name()))
        sio = StringIO()
        authzid = self.authzid
        if authzid is not None:
            sio.write(authzid)
        sio.write('\0')
        sio.write(self.authcid)
        sio.write('\0')
        sio.write(self.passwd)
        return sio.getvalue().encode(encoding)



from abc import ABCMeta, abstractmethod
from typing import Set

from ammoo.exceptions.connection import NoChallengesAuthenticationError


class AuthenticationMechanism(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def name(cls) -> str:
        raise NotImplementedError()

    @abstractmethod
    def build_start_ok_response(self, encoding: str, rabbitmq: bool) -> bytes:
        raise NotImplementedError()

    async def challenge(self, challenge: bytes) -> bytes:
        raise NoChallengesAuthenticationError(self.name())


class Authentication(metaclass=ABCMeta):
    @abstractmethod
    def select_mechanism(self, server_mechanisms: Set[str]) -> AuthenticationMechanism:
        raise NotImplementedError()



from asyncio import StreamReader
from collections import namedtuple
from datetime import datetime
from io import BytesIO
from struct import unpack, pack
from typing import NamedTuple, Optional

from ammoo.wire.classes import CLASS_BASIC
from ammoo.wire.low.containers import parse_field_table, STRING_ARGS, FIELD_ARGS, NO_ARGS, build_field_table, FieldTable
from ammoo.wire.low.exceptions import InvalidWeight, ExcessiveFramePayload, UnsupportedHeaderFrameClass, \
    TooManyHeaderProperties
from ammoo.wire.low.misc import parse_timestamp, parse_delivery_mode, build_timestamp
from ammoo.wire.low.strings import parse_shortstr, parse_binary_shortstr, build_shortstr, build_binary_shortstr
from ammoo.wire.low.structs import parse_short_uint, parse_octet_uint, pack_long_long_uint, pack_short_uint, \
    build_octet_uint, ZERO_SHORT_UINT_DATA
from ammoo.wire.typing import DeliveryMode


_BasicHeaderProperty = namedtuple('_BasicHeaderProperty', ['bit', 'name', 'parse', 'build', 'args_type'])

_BASIC_HEADER_PROPERTIES = [
    _BasicHeaderProperty(1 << 15, 'content_type', parse_shortstr, build_shortstr, STRING_ARGS),
    _BasicHeaderProperty(1 << 14, 'content_encoding', parse_shortstr, build_shortstr, STRING_ARGS),
    _BasicHeaderProperty(1 << 13, 'headers', parse_field_table, build_field_table, FIELD_ARGS),
    _BasicHeaderProperty(1 << 12, 'delivery_mode', parse_delivery_mode, build_octet_uint, NO_ARGS),
    _BasicHeaderProperty(1 << 11, 'priority', parse_octet_uint, build_octet_uint, NO_ARGS),
    _BasicHeaderProperty(1 << 10, 'correlation_id', parse_shortstr, build_shortstr, STRING_ARGS),
    _BasicHeaderProperty(1 << 9,  'reply_to', parse_shortstr, build_shortstr, STRING_ARGS),
    _BasicHeaderProperty(1 << 8,  'expiration', parse_shortstr, build_shortstr, STRING_ARGS),
    _BasicHeaderProperty(1 << 7,  'message_id', parse_shortstr, build_shortstr, STRING_ARGS),
    _BasicHeaderProperty(1 << 6,  'timestamp', parse_timestamp, build_timestamp, NO_ARGS),
    _BasicHeaderProperty(1 << 5,  'type_', parse_shortstr, build_shortstr, STRING_ARGS),
    _BasicHeaderProperty(1 << 4,  'user_id', parse_shortstr, build_shortstr, STRING_ARGS),
    _BasicHeaderProperty(1 << 3,  'app_id', parse_shortstr, build_shortstr, STRING_ARGS),
    # this is a rabbitmq extension, vanilla AMQP leaves it reserved
    # XXX: should this be non-binary?
    _BasicHeaderProperty(1 << 2,  'cluster_id', parse_binary_shortstr, build_binary_shortstr, NO_ARGS),
]


BasicHeaderProperties = NamedTuple('BasicHeaderProperties', [
    ('content_type', Optional[str]),
    ('content_encoding', Optional[str]),
    ('headers', Optional[FieldTable]),
    ('delivery_mode', Optional[DeliveryMode]),
    ('priority', Optional[int]),
    ('correlation_id', Optional[str]),
    ('reply_to', Optional[str]),  # RoutingKey?
    ('expiration', Optional[str]),
    ('message_id', Optional[str]),
    ('timestamp', Optional[datetime]),
    ('type_', Optional[str]),
    ('user_id', Optional[str]),
    ('app_id', Optional[str]),
    ('cluster_id', Optional[str]),
])


def parse_basic_header_properties(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicHeaderProperties:
    # The basic class only has 13 (+1 unused) properties, which means there will never be a second 16-bit value set
    # If there would, then this single loop would not be enough
    values = []

    flags = parse_short_uint(bio)
    if flags & 1 != 0:
        # Basic only has 13(+1) properties
        raise TooManyHeaderProperties(CLASS_BASIC)

    for property in _BASIC_HEADER_PROPERTIES:
        if flags & property.bit:
            if property.args_type == NO_ARGS:
                value = property.parse(bio)
            elif property.args_type == STRING_ARGS:
                value = property.parse(bio, encoding)
            else:
                value = property.parse(bio, encoding, rabbitmq)
            values.append(value)
        else:
            values.append(None)

    return BasicHeaderProperties(*values)


def build_basic_header_properties(
        bio: BytesIO, properties: BasicHeaderProperties, encoding: str, rabbitmq: bool) -> None:
    flags = 0
    properties_bio = BytesIO()
    for property, value in zip(_BASIC_HEADER_PROPERTIES, properties):
        if value is not None:
            flags |= property.bit
            if property.args_type == NO_ARGS:
                property.build(properties_bio, value)
            elif property.args_type == STRING_ARGS:
                property.build(properties_bio, value, encoding)
            else:
                property.build(properties_bio, value, encoding, rabbitmq)
    bio.write(pack_short_uint(flags))
    bio.write(properties_bio.getvalue())


def build_empty_header_properties(bio: BytesIO) -> None:
    """When all properties are None, this just writes a flags int with all false"""
    bio.write(ZERO_SHORT_UINT_DATA)


FRAME_TYPE_HEADER = 2


class BasicHeaderFrame:
    __slots__ = 'channel_id', 'body_size', 'properties'
    frame_type = FRAME_TYPE_HEADER

    def __init__(self, channel_id, body_size, properties):
        self.channel_id = channel_id
        self.body_size = body_size
        self.properties = properties

    def __str__(self):
        return 'Header frame for class basic, body size {}, properties {}'.format(self.body_size, self.properties)


async def parse_header_frame(
        reader: StreamReader, payload_size: int, channel_id: int, encoding: str, rabbitmq: bool) -> BasicHeaderFrame:
    class_id, weight, body_size = unpack('!HHQ', await reader.readexactly(12))
    if class_id != CLASS_BASIC:
        raise UnsupportedHeaderFrameClass(class_id)
    if weight != 0:
        raise InvalidWeight(weight)
    bio = BytesIO(await reader.readexactly(payload_size - 12))
    properties = parse_basic_header_properties(bio, encoding, rabbitmq)
    payload_offset = bio.tell() + 12
    if payload_offset != payload_size:
        raise ExcessiveFramePayload(payload_size, payload_offset)
    return BasicHeaderFrame(channel_id, body_size, properties)


BASIC_HEADER_PAYLOAD_PREFIX = pack('!HH', CLASS_BASIC, 0)  # class and weight


def pack_basic_header_frame_payload(
        body_size: int, properties: Optional[BasicHeaderProperties], encoding: str, rabbitmq: bool) -> bytes:
    bio = BytesIO()
    bio.write(BASIC_HEADER_PAYLOAD_PREFIX)
    bio.write(pack_long_long_uint(body_size))
    if properties is None:
        build_empty_header_properties(bio)
    else:
        build_basic_header_properties(bio, properties, encoding, rabbitmq)
    return bio.getvalue()
